""" Twins
A PyTorch impl of : `Twins: Revisiting the Design of Spatial Attention in Vision Transformers`
    - https://arxiv.org/pdf/2104.13840.pdf
Code/weights from https://github.com/Meituan-AutoML/Twins, original copyright/license info below
"""

# --------------------------------------------------------
# Twins
# Copyright (c) 2021 Meituan
# Licensed under The Apache 2.0 License [see LICENSE for details]
# Written by Xinjie Li, Xiangxiang Chu
# --------------------------------------------------------
import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial

from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.layers import Mlp, DropPath, to_2tuple, trunc_normal_
from timm.models.vision_transformer import Attention
from .attention import LinearPositionEmbeddingSine
from .utils import coords_grid


def _cfg(url="", **kwargs):
    return {
        "url": url,
        "num_classes": 1000,
        "input_size": (3, 224, 224),
        "pool_size": None,
        "crop_pct": 0.9,
        "interpolation": "bicubic",
        "fixed_input_size": True,
        "mean": IMAGENET_DEFAULT_MEAN,
        "std": IMAGENET_DEFAULT_STD,
        "first_conv": "patch_embeds.0.proj",
        "classifier": "head",
        **kwargs,
    }


default_cfgs = {
    "twins_pcpvt_small": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_small-e70e7e7a.pth",
    ),
    "twins_pcpvt_base": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_base-e5ecb09b.pth",
    ),
    "twins_pcpvt_large": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_pcpvt_large-d273f802.pth",
    ),
    "twins_svt_small": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_small-42e5f78c.pth",
    ),
    "twins_svt_base": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_base-c2265010.pth",
    ),
    "twins_svt_large": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vt3p-weights/twins_svt_large-90f6aaa9.pth",
    ),
}

Size_ = Tuple[int, int]


class GroupAttnRPEContext(nn.Module):
    """Latent cost tokens attend to different group"""

    def __init__(
        self,
        dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        ws=1,
        cfg=None,
        vert_c_dim=0,
    ):
        super(GroupAttnRPEContext, self).__init__()
        assert ws != 1
        assert cfg is not None
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."
        assert (
            cfg.cost_latent_token_num % 5 == 0
        ), "cost_latent_token_num should be divided by 5."
        assert vert_c_dim > 0, "vert_c_dim should not be 0"

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5
        self.vert_c_dim = vert_c_dim

        self.cfg = cfg

        self.context_proj = nn.Linear(256, vert_c_dim)
        self.q = nn.Linear(dim + vert_c_dim, dim, bias=True)
        self.k = nn.Linear(dim + vert_c_dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_, context=None):
        B, N, C = x.shape
        C_qk = C + self.vert_c_dim
        H, W = size
        batch_num = B // 5

        context = context.repeat(B // context.shape[0], 1, 1, 1)
        context = context.view(B, -1, H * W).permute(0, 2, 1)
        context = self.context_proj(context)
        context = context.view(B, H, W, -1)

        x = x.view(B, H, W, C)
        x_qk = torch.cat([x, context], dim=-1)

        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        x_qk = F.pad(x_qk, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        padded_N = Hp * Wp

        coords = coords_grid(B, Hp, Wp, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C_qk)
        coords_enc = coords_enc.reshape(B, Hp, Wp, C_qk)

        q = (
            self.q(x_qk + coords_enc)
            .reshape(B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads)
            .transpose(2, 3)
        )
        q = q.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        v = self.v(x)
        k = self.k(x_qk + coords_enc)
        # concate and do shifting operation together
        kv = torch.cat([k, v], dim=-1)
        kv_up = torch.cat(
            [
                kv[:batch_num, self.ws : Hp, :, :],
                kv[:batch_num, Hp - self.ws : Hp, :, :],
            ],
            dim=1,
        )
        kv_down = torch.cat(
            [
                kv[batch_num : batch_num * 2, : self.ws, :, :],
                kv[batch_num : batch_num * 2, : Hp - self.ws, :, :],
            ],
            dim=1,
        )
        kv_left = torch.cat(
            [
                kv[batch_num * 2 : batch_num * 3, :, self.ws : Wp, :],
                kv[batch_num * 2 : batch_num * 3, :, Wp - self.ws : Wp, :],
            ],
            dim=2,
        )
        kv_right = torch.cat(
            [
                kv[batch_num * 3 : batch_num * 4, :, : self.ws, :],
                kv[batch_num * 3 : batch_num * 4, :, : Wp - self.ws, :],
            ],
            dim=2,
        )
        kv_center = kv[batch_num * 4 : batch_num * 5, :, :, :]
        kv_shifted = torch.cat([kv_up, kv_down, kv_left, kv_right, kv_center], dim=0)
        k, v = torch.split(kv_shifted, [self.dim, self.dim], dim=-1)

        k = k.reshape(
            B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads
        ).transpose(2, 3)
        k = k.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        v = v.reshape(
            B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads
        ).transpose(2, 3)
        v = v.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class GroupAttnRPE(nn.Module):
    """Latent cost tokens attend to different group"""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, ws=1, cfg=None):
        super(GroupAttnRPE, self).__init__()
        assert ws != 1
        assert cfg is not None
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."
        assert (
            cfg.cost_latent_token_num % 5 == 0
        ), "cost_latent_token_num should be divided by 5."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.cfg = cfg

        self.q = nn.Linear(dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_, context=None):
        B, N, C = x.shape
        H, W = size
        batch_num = B // 5
        x = x.view(B, H, W, C)
        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        padded_N = Hp * Wp

        coords = coords_grid(B, Hp, Wp, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        coords_enc = coords_enc.reshape(B, Hp, Wp, C)

        q = (
            self.q(x + coords_enc)
            .reshape(B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads)
            .transpose(2, 3)
        )
        q = q.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        v = self.v(x)
        k = self.k(x + coords_enc)
        # concate and do shifting operation together
        kv = torch.cat([k, v], dim=-1)
        kv_up = torch.cat(
            [
                kv[:batch_num, self.ws : Hp, :, :],
                kv[:batch_num, Hp - self.ws : Hp, :, :],
            ],
            dim=1,
        )
        kv_down = torch.cat(
            [
                kv[batch_num : batch_num * 2, : self.ws, :, :],
                kv[batch_num : batch_num * 2, : Hp - self.ws, :, :],
            ],
            dim=1,
        )
        kv_left = torch.cat(
            [
                kv[batch_num * 2 : batch_num * 3, :, self.ws : Wp, :],
                kv[batch_num * 2 : batch_num * 3, :, Wp - self.ws : Wp, :],
            ],
            dim=2,
        )
        kv_right = torch.cat(
            [
                kv[batch_num * 3 : batch_num * 4, :, : self.ws, :],
                kv[batch_num * 3 : batch_num * 4, :, : Wp - self.ws, :],
            ],
            dim=2,
        )
        kv_center = kv[batch_num * 4 : batch_num * 5, :, :, :]
        kv_shifted = torch.cat([kv_up, kv_down, kv_left, kv_right, kv_center], dim=0)
        k, v = torch.split(kv_shifted, [self.dim, self.dim], dim=-1)

        k = k.reshape(
            B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads
        ).transpose(2, 3)
        k = k.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        v = v.reshape(
            B, _h, self.ws, _w, self.ws, self.num_heads, C // self.num_heads
        ).transpose(2, 3)
        v = v.reshape(
            B, _h * _w, self.ws * self.ws, self.num_heads, C // self.num_heads
        ).permute(0, 1, 3, 2, 4)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class LocallyGroupedAttnRPEContext(nn.Module):
    """LSA: self attention within a group"""

    def __init__(
        self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, ws=1, vert_c_dim=0
    ):
        assert ws != 1
        super(LocallyGroupedAttnRPEContext, self).__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5
        self.vert_c_dim = vert_c_dim

        self.context_proj = nn.Linear(256, vert_c_dim)
        # context are not added to value
        self.q = nn.Linear(dim + vert_c_dim, dim, bias=True)
        self.k = nn.Linear(dim + vert_c_dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_, context=None):
        # There are two implementations for this function, zero padding or mask. We don't observe obvious difference for
        # both. You can choose any one, we recommend forward_padding because it's neat. However,
        # the masking implementation is more reasonable and accurate.
        B, N, C = x.shape
        H, W = size
        C_qk = C + self.vert_c_dim

        context = context.repeat(B // context.shape[0], 1, 1, 1)
        context = context.view(B, -1, H * W).permute(0, 2, 1)
        context = self.context_proj(context)
        context = context.view(B, H, W, -1)

        x = x.view(B, H, W, C)
        x_qk = torch.cat([x, context], dim=-1)

        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        x_qk = F.pad(x_qk, (0, 0, pad_l, pad_r, pad_t, pad_b))

        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        x = x.reshape(B, _h, self.ws, _w, self.ws, C).transpose(2, 3)
        x_qk = x_qk.reshape(B, _h, self.ws, _w, self.ws, C_qk).transpose(2, 3)

        v = (
            self.v(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )

        coords = coords_grid(B, self.ws, self.ws, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C_qk).view(
            B, self.ws, self.ws, C_qk
        )
        # coords_enc:   B, ws, ws, C
        # x:            B, _h, _w, self.ws, self.ws, C
        x_qk = x_qk + coords_enc[:, None, None, :, :, :]

        q = (
            self.q(x_qk)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )
        k = (
            self.k(x_qk)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class GlobalSubSampleAttnRPEContext(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(
        self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1, vert_c_dim=0
    ):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.vert_c_dim = vert_c_dim
        self.context_proj = nn.Linear(256, vert_c_dim)
        self.q = nn.Linear(dim + vert_c_dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr_key = nn.Conv2d(
                dim + vert_c_dim, dim, kernel_size=sr_ratio, stride=sr_ratio
            )
            self.sr_value = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, size: Size_, context=None):
        B, N, C = x.shape
        C_qk = C + self.vert_c_dim
        H, W = size
        context = context.repeat(B // context.shape[0], 1, 1, 1)
        context = context.view(B, -1, H * W).permute(0, 2, 1)
        context = self.context_proj(context)
        context = context.view(B, H, W, -1)
        x = x.view(B, H, W, C)
        x_qk = torch.cat([x, context], dim=-1)
        pad_l = pad_t = 0
        pad_r = (self.sr_ratio - W % self.sr_ratio) % self.sr_ratio
        pad_b = (self.sr_ratio - H % self.sr_ratio) % self.sr_ratio
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        x_qk = F.pad(x_qk, (0, 0, pad_l, pad_r, pad_t, pad_b))

        _, Hp, Wp, _ = x.shape
        padded_size = (Hp, Wp)
        padded_N = Hp * Wp
        x = x.view(B, -1, C)
        x_qk = x_qk.view(B, -1, C_qk)

        coords = coords_grid(B, *padded_size, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C_qk)
        # coords_enc:   B, Hp*Wp, C
        # x:            B, Hp*Wp, C
        q = (
            self.q(x_qk + coords_enc)
            .reshape(B, padded_N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr_key is not None:
            x = x.permute(0, 2, 1).reshape(B, C, *padded_size)
            x_qk = x_qk.permute(0, 2, 1).reshape(B, C_qk, *padded_size)
            x = self.sr_value(x).reshape(B, C, -1).permute(0, 2, 1)
            x_qk = self.sr_key(x_qk).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)
            x_qk = self.norm(x_qk)

        coords = coords_grid(
            B,
            padded_size[0] // self.sr_ratio,
            padded_size[1] // self.sr_ratio,
            dtype=x.dtype,
            device=x.device,
        )
        coords = coords.view(B, 2, -1).permute(0, 2, 1) * self.sr_ratio
        # align the coordinate of local and global
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        k = (
            self.k(x_qk + coords_enc)
            .reshape(
                B,
                (padded_size[0] // self.sr_ratio) * (padded_size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )
        v = (
            self.v(x)
            .reshape(
                B,
                (padded_size[0] // self.sr_ratio) * (padded_size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, Hp, Wp, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()

        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class LocallyGroupedAttnRPE(nn.Module):
    """LSA: self attention within a group"""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, ws=1):
        assert ws != 1
        super(LocallyGroupedAttnRPE, self).__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_, context=None):
        # There are two implementations for this function, zero padding or mask. We don't observe obvious difference for
        # both. You can choose any one, we recommend forward_padding because it's neat. However,
        # the masking implementation is more reasonable and accurate.
        B, N, C = x.shape
        H, W = size
        x = x.view(B, H, W, C)
        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        x = x.reshape(B, _h, self.ws, _w, self.ws, C).transpose(2, 3)
        v = (
            self.v(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )

        coords = coords_grid(B, self.ws, self.ws, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C).view(
            B, self.ws, self.ws, C
        )
        # coords_enc:   B, ws, ws, C
        # x:            B, _h, _w, self.ws, self.ws, C
        x = x + coords_enc[:, None, None, :, :, :]

        q = (
            self.q(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )
        k = (
            self.k(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 1, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)[0]
        )
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class GlobalSubSampleAttnRPE(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, size: Size_, context=None):
        B, N, C = x.shape
        H, W = size
        x = x.view(B, H, W, C)
        pad_l = pad_t = 0
        pad_r = (self.sr_ratio - W % self.sr_ratio) % self.sr_ratio
        pad_b = (self.sr_ratio - H % self.sr_ratio) % self.sr_ratio
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        padded_size = (Hp, Wp)
        padded_N = Hp * Wp
        x = x.view(B, -1, C)

        coords = coords_grid(B, *padded_size, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        # coords_enc:   B, Hp*Wp, C
        # x:            B, Hp*Wp, C
        q = (
            self.q(x + coords_enc)
            .reshape(B, padded_N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr is not None:
            x = x.permute(0, 2, 1).reshape(B, C, *padded_size)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)

        coords = coords_grid(
            B,
            padded_size[0] // self.sr_ratio,
            padded_size[1] // self.sr_ratio,
            dtype=x.dtype,
            device=x.device,
        )
        coords = coords.view(B, 2, -1).permute(0, 2, 1) * self.sr_ratio
        # align the coordinate of local and global
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        k = (
            self.k(x + coords_enc)
            .reshape(
                B,
                (padded_size[0] // self.sr_ratio) * (padded_size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )
        v = (
            self.v(x)
            .reshape(
                B,
                (padded_size[0] // self.sr_ratio) * (padded_size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, Hp, Wp, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()

        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class CrossGlobalSubSampleAttnRPE(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, tgt, size: Size_):
        B, N, C = x.shape
        coords = coords_grid(B, *size, dtype=x.dtype, device=x.device)
        coords = coords.view(B, 2, -1).permute(0, 2, 1)
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        # coords_enc:   B, H*W, C
        # x:            B, H*W, C
        q = (
            self.q(x + coords_enc)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr is not None:
            tgt = tgt.permute(0, 2, 1).reshape(B, C, *size)
            tgt = self.sr(tgt).reshape(B, C, -1).permute(0, 2, 1)
            tgt = self.norm(tgt)
        coords = coords_grid(
            B,
            size[0] // self.sr_ratio,
            size[1] // self.sr_ratio,
            dtype=x.dtype,
            device=x.device,
        )
        coords = coords.view(B, 2, -1).permute(0, 2, 1) * self.sr_ratio
        # align the coordinate of local and global
        coords_enc = LinearPositionEmbeddingSine(coords, dim=C)
        k = (
            self.k(tgt + coords_enc)
            .reshape(
                B,
                (size[0] // self.sr_ratio) * (size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )
        v = (
            self.v(tgt)
            .reshape(
                B,
                (size[0] // self.sr_ratio) * (size[1] // self.sr_ratio),
                self.num_heads,
                C // self.num_heads,
            )
            .permute(0, 2, 1, 3)
        )

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class LocallyGroupedAttn(nn.Module):
    """LSA: self attention within a group"""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, ws=1):
        assert ws != 1
        super(LocallyGroupedAttn, self).__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.ws = ws

    def forward(self, x, size: Size_):
        # There are two implementations for this function, zero padding or mask. We don't observe obvious difference for
        # both. You can choose any one, we recommend forward_padding because it's neat. However,
        # the masking implementation is more reasonable and accurate.
        B, N, C = x.shape
        H, W = size
        x = x.view(B, H, W, C)
        pad_l = pad_t = 0
        pad_r = (self.ws - W % self.ws) % self.ws
        pad_b = (self.ws - H % self.ws) % self.ws
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape
        _h, _w = Hp // self.ws, Wp // self.ws
        x = x.reshape(B, _h, self.ws, _w, self.ws, C).transpose(2, 3)
        qkv = (
            self.qkv(x)
            .reshape(
                B, _h * _w, self.ws * self.ws, 3, self.num_heads, C // self.num_heads
            )
            .permute(3, 0, 1, 4, 2, 5)
        )
        q, k, v = qkv[0], qkv[1], qkv[2]
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        attn = (attn @ v).transpose(2, 3).reshape(B, _h, _w, self.ws, self.ws, C)
        x = attn.transpose(2, 3).reshape(B, _h * self.ws, _w * self.ws, C)
        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()
        x = x.reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class GlobalSubSampleAttn(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.kv = nn.Linear(dim, dim * 2, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, size: Size_):
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr is not None:
            x = x.permute(0, 2, 1).reshape(B, C, *size)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)
        kv = (
            self.kv(x)
            .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class CrossGlobalSubSampleAttn(nn.Module):
    """GSA: using a  key to summarize the information for a group to be efficient."""

    def __init__(self, dim, num_heads=8, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=True)
        self.kv = nn.Linear(dim, dim * 2, bias=True)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)
        else:
            self.sr = None
            self.norm = None

    def forward(self, x, tgt, size: Size_):
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr is not None:
            tgt = tgt.permute(0, 2, 1).reshape(B, C, *size)
            tgt = self.sr(tgt).reshape(B, C, -1).permute(0, 2, 1)
            tgt = self.norm(tgt)
        kv = (
            self.kv(tgt)
            .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class CrossBlock(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio=1,
        ws=None,
        with_rpe=True,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = CrossGlobalSubSampleAttnRPE(
            dim, num_heads, attn_drop, drop, sr_ratio
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, src, tgt, size: Size_):
        src_shortcut, tgt_shortcut = src, tgt

        src, tgt = self.norm1(src), self.norm1(tgt)
        src = src_shortcut + self.drop_path(self.attn(src, tgt, size))
        tgt = tgt_shortcut + self.drop_path(self.attn(tgt, src, size))

        src = src + self.drop_path(self.mlp(self.norm2(src)))
        tgt = tgt + self.drop_path(self.mlp(self.norm2(tgt)))
        return src, tgt


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio=1,
        ws=None,
        with_rpe=False,
        vert_c_dim=0,
        groupattention=False,
        cfg=None,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        if groupattention:
            assert with_rpe, "Not implementing groupattention without rpe"
            if vert_c_dim > 0:
                self.attn = GroupAttnRPEContext(
                    dim, num_heads, attn_drop, drop, ws, cfg, vert_c_dim
                )
            else:
                self.attn = GroupAttnRPE(dim, num_heads, attn_drop, drop, ws, cfg)
        elif ws is None:
            self.attn = Attention(dim, num_heads, False, None, attn_drop, drop)
        elif ws == 1:
            if with_rpe:
                if vert_c_dim > 0:
                    self.attn = GlobalSubSampleAttnRPEContext(
                        dim, num_heads, attn_drop, drop, sr_ratio, vert_c_dim
                    )
                else:
                    self.attn = GlobalSubSampleAttnRPE(
                        dim, num_heads, attn_drop, drop, sr_ratio
                    )
            else:
                self.attn = GlobalSubSampleAttn(
                    dim, num_heads, attn_drop, drop, sr_ratio
                )
        else:
            if with_rpe:
                if vert_c_dim > 0:
                    self.attn = LocallyGroupedAttnRPEContext(
                        dim, num_heads, attn_drop, drop, ws, vert_c_dim
                    )
                else:
                    self.attn = LocallyGroupedAttnRPE(
                        dim, num_heads, attn_drop, drop, ws
                    )
            else:
                self.attn = LocallyGroupedAttn(dim, num_heads, attn_drop, drop, ws)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, size: Size_, context=None):
        x = x + self.drop_path(self.attn(self.norm1(x), size, context))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PosConv(nn.Module):
    # PEG  from https://arxiv.org/abs/2102.10882
    def __init__(self, in_chans, embed_dim=768, stride=1):
        super(PosConv, self).__init__()
        self.proj = nn.Sequential(
            nn.Conv2d(in_chans, embed_dim, 3, stride, 1, bias=True, groups=embed_dim),
        )
        self.stride = stride

    def forward(self, x, size: Size_):
        B, N, C = x.shape
        cnn_feat_token = x.transpose(1, 2).view(B, C, *size)
        x = self.proj(cnn_feat_token)
        if self.stride == 1:
            x += cnn_feat_token
        x = x.flatten(2).transpose(1, 2)
        return x

    def no_weight_decay(self):
        return ["proj.%d.weight" % i for i in range(4)]


class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        assert (
            img_size[0] % patch_size[0] == 0 and img_size[1] % patch_size[1] == 0
        ), f"img_size {img_size} should be divided by patch_size {patch_size}."
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x) -> Tuple[torch.Tensor, Size_]:
        B, C, H, W = x.shape

        x = self.proj(x).flatten(2).transpose(1, 2)
        x = self.norm(x)
        out_size = (H // self.patch_size[0], W // self.patch_size[1])

        return x, out_size


class Twins(nn.Module):
    """Twins Vision Transfomer (Revisiting Spatial Attention)
    Adapted from PVT (PyramidVisionTransformer) class at https://github.com/whai362/PVT.git
    """

    def __init__(
        self,
        img_size=224,
        patch_size=4,
        in_chans=3,
        num_classes=1000,
        embed_dims=(64, 128, 256, 512),
        num_heads=(1, 2, 4, 8),
        mlp_ratios=(4, 4, 4, 4),
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=(3, 4, 6, 3),
        sr_ratios=(8, 4, 2, 1),
        wss=None,
        block_cls=Block,
        init_weight=True,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths
        self.embed_dims = embed_dims
        self.num_features = embed_dims[-1]

        img_size = to_2tuple(img_size)
        prev_chs = in_chans
        self.patch_embeds = nn.ModuleList()
        self.pos_drops = nn.ModuleList()
        for i in range(len(depths)):
            self.patch_embeds.append(
                PatchEmbed(img_size, patch_size, prev_chs, embed_dims[i])
            )
            self.pos_drops.append(nn.Dropout(p=drop_rate))
            prev_chs = embed_dims[i]
            img_size = tuple(t // patch_size for t in img_size)
            patch_size = 2

        self.blocks = nn.ModuleList()
        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule
        cur = 0
        for k in range(len(depths)):
            _block = nn.ModuleList(
                [
                    block_cls(
                        dim=embed_dims[k],
                        num_heads=num_heads[k],
                        mlp_ratio=mlp_ratios[k],
                        drop=drop_rate,
                        attn_drop=attn_drop_rate,
                        drop_path=dpr[cur + i],
                        norm_layer=norm_layer,
                        sr_ratio=sr_ratios[k],
                        ws=1 if wss is None or i % 2 == 1 else wss[k],
                    )
                    for i in range(depths[k])
                ]
            )
            self.blocks.append(_block)
            cur += depths[k]

        self.pos_block = nn.ModuleList(
            [PosConv(embed_dim, embed_dim) for embed_dim in embed_dims]
        )

        self.norm = norm_layer(self.num_features)

        # classification head
        self.head = (
            nn.Linear(self.num_features, num_classes)
            if num_classes > 0
            else nn.Identity()
        )

        # init weights
        if init_weight:
            self.apply(self._init_weights)

    @torch.jit.ignore
    def no_weight_decay(self):
        return set(["pos_block." + n for n, p in self.pos_block.named_parameters()])

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=""):
        self.num_classes = num_classes
        self.head = (
            nn.Linear(self.num_features, num_classes)
            if num_classes > 0
            else nn.Identity()
        )

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1.0)
            m.bias.data.zero_()

    def forward_features(self, x):
        B = x.shape[0]
        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(self.patch_embeds, self.pos_drops, self.blocks, self.pos_block)
        ):
            x, size = embed(x)
            x = drop(x)
            for j, blk in enumerate(blocks):
                x = blk(x, size)
                if j == 0:
                    x = pos_blk(x, size)  # PEG here
            if i < len(self.depths) - 1:
                x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()
        x = self.norm(x)
        return x.mean(dim=1)  # GAP here

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)
        return x



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .cgu import CGUStage
from .local_timm.weight_init import trunc_normal_
from .res_stem import ResStem
from .conv import Conv2dBlock, ConvTranspose2dBlock
from .norm import GroupNorm2d
from .update import ConvGRU


class CGUBidirDualEncoder(nn.Module):
    def __init__(
        self,
        pyramid_levels: Optional[int],
        hidden_chs: int,
        out_1x1_abs_chs: int,
        out_1x1_factor: Optional[float],
        num_out_stages: int = 0,
        gru_mode: str = "gru",
        activation_function: Optional[nn.Module] = None,
        norm_layer=GroupNorm2d,
        depth: int = 2,
        mlp_ratio: float = 4,
        mlp_use_dw_conv: bool = False,
        mlp_dw_kernel_size: int = 7,
        mlp_in_kernel_size: int = 1,
        mlp_out_kernel_size: int = 1,
        cgu_layer_scale_init_value: float = 1e-2,
    ):
        super().__init__()

        self.pyramid_levels = pyramid_levels
        self.hidden_chs = hidden_chs
        self.out_1x1_abs_chs = out_1x1_abs_chs
        self.out_1x1_factor = out_1x1_factor
        self.num_out_stages = num_out_stages
        self.gru_mode = gru_mode

        self.forward_gru = ConvGRU(hidden_chs[-1], hidden_chs[-1])
        self.down_gru = Conv2dBlock(
            hidden_chs[-1], hidden_chs[-1], 3, stride=2, padding=1, bias=True
        )

        self.backward_gru = ConvGRU(hidden_chs[-1], hidden_chs[-1])
        self.up_gru = ConvTranspose2dBlock(
            hidden_chs[-1], hidden_chs[-1], 4, stride=2, padding=1, bias=True
        )

        self.stem = self._make_stem(
            [hidden_chs[0], hidden_chs[1], 2 * hidden_chs[2]],
            norm_layer=norm_layer,
        )

        self.lowres_stem = self._make_stem(
            hidden_chs,
            norm_layer=norm_layer,
        )

        if self.out_1x1_abs_chs > 0:
            self.out_1x1 = self._make_out_1x1_layer(
                hidden_chs[-1], self.out_1x1_abs_chs
            )

        self.rec_stage = self._make_stage(
            hidden_chs[-1],
            out_chs=hidden_chs[-1],
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            cgu_layer_scale_init_value=cgu_layer_scale_init_value,
        )

        self.back_stage = self._make_stage(
            hidden_chs[-1],
            out_chs=hidden_chs[-1],
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            stride=1,
            cgu_layer_scale_init_value=cgu_layer_scale_init_value,
        )

        if self.num_out_stages > 0:
            self.out_merge_conv = Conv2dBlock(
                3 * hidden_chs[-1], hidden_chs[-1], kernel_size=1, pre_act_fn=nn.ReLU()
            )
            self.out_stages = self._make_out_stages(
                self.num_out_stages,
                hidden_chs[-1],
                out_chs=None,
                activation_function=activation_function,
                norm_layer=norm_layer,
                depth=depth,
                mlp_ratio=mlp_ratio,
                mlp_use_dw_conv=mlp_use_dw_conv,
                mlp_dw_kernel_size=mlp_dw_kernel_size,
                mlp_in_kernel_size=mlp_in_kernel_size,
                mlp_out_kernel_size=mlp_out_kernel_size,
                cgu_layer_scale_init_value=cgu_layer_scale_init_value,
            )

        self._init_weights()

    def _make_stem(self, hidden_chs: int, norm_layer):
        if not isinstance(hidden_chs, (list, tuple)):
            return

        return ResStem([hidden_chs[0], hidden_chs[1], hidden_chs[2]], norm_layer)

    def _make_stage(
        self,
        hidden_chs: int,
        out_chs=None,
        activation_function=None,
        norm_layer=GroupNorm2d,
        depth=2,
        mlp_ratio=4,
        mlp_use_dw_conv=True,
        mlp_dw_kernel_size=7,
        mlp_in_kernel_size=1,
        mlp_out_kernel_size=1,
        stride=2,
        cgu_layer_scale_init_value=1e-2,
    ):
        if out_chs is None:
            out_chs = hidden_chs

        return CGUStage(
            hidden_chs,
            out_chs,
            stride=stride,
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=depth,
            use_cross=True,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            layer_scale_init_value=cgu_layer_scale_init_value,
        )

    def _make_out_stages(
        self,
        num_out_stages: int,
        hidden_chs: int,
        out_chs=None,
        activation_function=None,
        norm_layer=GroupNorm2d,
        depth=2,
        mlp_dw_kernel_size=7,
        mlp_ratio=4,
        mlp_use_dw_conv=True,
        mlp_in_kernel_size=1,
        mlp_out_kernel_size=1,
        cgu_layer_scale_init_value=1e-2,
    ):
        if out_chs is None:
            out_chs = hidden_chs
        return CGUStage(
            hidden_chs,
            out_chs,
            stride=1,
            activation_function=activation_function,
            norm_layer=norm_layer,
            depth=num_out_stages * depth,
            use_cross=True,
            mlp_ratio=mlp_ratio,
            mlp_use_dw_conv=mlp_use_dw_conv,
            mlp_dw_kernel_size=mlp_dw_kernel_size,
            mlp_in_kernel_size=mlp_in_kernel_size,
            mlp_out_kernel_size=mlp_out_kernel_size,
            layer_scale_init_value=cgu_layer_scale_init_value,
        )

    def _make_out_1x1_layer(self, hidden_chs: int, out_chs: int):
        return Conv2dBlock(hidden_chs, out_chs, kernel_size=1)

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=0.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1.0)
            elif isinstance(m, nn.Conv2d):
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    m.bias.data.zero_()

    def forward(self, x: torch.Tensor, y: torch.Tensor, pyr_levels: int):
        input_x = x
        input_y = y

        x_pyramid = []
        y_pyramid = []

        pyr_iters = pyr_levels + 1
        for i in range(pyr_iters):
            if i == 0:
                x = self.stem(x)
                y = self.stem(y)

                x, hx = torch.split(x, [x.shape[1] // 2, x.shape[1] // 2], 1)
                y, hy = torch.split(y, [y.shape[1] // 2, y.shape[1] // 2], 1)
                hx = torch.tanh(hx)
                hy = torch.tanh(hy)
            else:
                hx = self.forward_gru(hx, x)
                hy = self.forward_gru(hy, y)

                x, y = self.rec_stage(hx, hy)
                x = x.contiguous()
                y = y.contiguous()
                if i < (pyr_iters - 1):
                    hx = self.down_gru(hx)
                    hx = torch.tanh(hx)
                    hy = self.down_gru(hy)
                    hy = torch.tanh(hy)

            if i >= 1:
                x_pyramid.append(x)
                y_pyramid.append(y)

        hx = torch.zeros_like(x_pyramid[-1])
        hy = torch.zeros_like(y_pyramid[-1])
        for i in range(len(x_pyramid) - 1, -1, -1):
            x = x_pyramid[i]
            y = y_pyramid[i]

            hx = self.backward_gru(hx, x)
            hy = self.backward_gru(hy, y)

            x2, y2 = self.back_stage(hx, hy)

            input_x_lowres = F.interpolate(
                input_x,
                scale_factor=(1.0 / 2.0 ** (i + 1)),
                mode="bilinear",
                align_corners=True,
            )
            x_lowres = self.lowres_stem(input_x_lowres)

            input_y_lowres = F.interpolate(
                input_y,
                scale_factor=(1.0 / 2.0 ** (i + 1)),
                mode="bilinear",
                align_corners=True,
            )
            y_lowres = self.lowres_stem(input_y_lowres)

            x_pyramid[i] = torch.cat([x, x2, x_lowres], 1)
            y_pyramid[i] = torch.cat([y, y2, y_lowres], 1)

            if i > 0:
                hx = self.up_gru(hx)
                hx = torch.tanh(hx)
                hy = self.up_gru(hy)
                hy = torch.tanh(hy)

        for i, (x, y) in enumerate(zip(x_pyramid, y_pyramid)):
            if self.num_out_stages > 0:
                x = self.out_merge_conv(x)
                y = self.out_merge_conv(y)
                x, y = self.out_stages(x, y)
            if self.out_1x1_abs_chs > 0:
                if self.out_1x1_factor is None:
                    x = self.out_1x1(x)
                    y = self.out_1x1(y)
                else:
                    x = self.out_1x1(x, int(self.out_1x1_factor * x.shape[1]))
                    y = self.out_1x1(y, int(self.out_1x1_factor * y.shape[1]))
            x_pyramid[i] = x
            y_pyramid[i] = y

        return x_pyramid[::-1], y_pyramid[::-1]



import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.layers import trunc_normal_
import math


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(ResidualBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BottleneckBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(BottleneckBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_planes, planes // 4, kernel_size=1, padding=0)
        self.conv2 = nn.Conv2d(
            planes // 4, planes // 4, kernel_size=3, padding=1, stride=stride
        )
        self.conv3 = nn.Conv2d(planes // 4, planes, kernel_size=1, padding=0)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm4 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes // 4)
            self.norm2 = nn.BatchNorm2d(planes // 4)
            self.norm3 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes // 4)
            self.norm2 = nn.InstanceNorm2d(planes // 4)
            self.norm3 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()
            if not stride == 1:
                self.norm4 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm4
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))
        y = self.relu(self.norm3(self.conv3(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BasicEncoder(nn.Module):
    def __init__(self, input_dim=3, output_dim=128, norm_fn="batch", dropout=0.0):
        super(BasicEncoder, self).__init__()
        self.norm_fn = norm_fn
        mul = input_dim // 3

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64 * mul)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64 * mul)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64 * mul)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(input_dim, 64 * mul, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64 * mul
        self.layer1 = self._make_layer(64 * mul, stride=1)
        self.layer2 = self._make_layer(96 * mul, stride=2)
        self.layer3 = self._make_layer(128 * mul, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128 * mul, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class SmallEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(SmallEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=32)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(32)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(32)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 32, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 32
        self.layer1 = self._make_layer(32, stride=1)
        self.layer2 = self._make_layer(64, stride=2)
        self.layer3 = self._make_layer(96, stride=2)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        self.conv2 = nn.Conv2d(96, output_dim, kernel_size=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = BottleneckBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = BottleneckBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class ConvNets(nn.Module):
    def __init__(self, in_dim, out_dim, inter_dim, depth, stride=1):
        super(ConvNets, self).__init__()

        self.conv_first = nn.Conv2d(
            in_dim, inter_dim, kernel_size=3, padding=1, stride=stride
        )
        self.conv_last = nn.Conv2d(
            inter_dim, out_dim, kernel_size=3, padding=1, stride=stride
        )
        self.relu = nn.ReLU(inplace=True)
        self.inter_convs = nn.ModuleList(
            [
                ResidualBlock(inter_dim, inter_dim, norm_fn="none", stride=1)
                for i in range(depth)
            ]
        )

    def forward(self, x):
        x = self.relu(self.conv_first(x))
        for inter_conv in self.inter_convs:
            x = inter_conv(x)
        x = self.conv_last(x)
        return x


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class ConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(ConvGRU, self).__init__()
        self.convz = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convr = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)
        self.convq = nn.Conv2d(hidden_dim + input_dim, hidden_dim, 3, padding=1)

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = torch.sigmoid(self.convz(hx))
        r = torch.sigmoid(self.convr(hx))
        q = torch.tanh(self.convq(torch.cat([r * h, x], dim=1)))

        h = (1 - z) * h + z * q
        return h


class SepConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convr1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convq1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )

        self.convz2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convr2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convq2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )

    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        return h


class BasicMotionEncoder(nn.Module):
    def __init__(self, args):
        super(BasicMotionEncoder, self).__init__()
        cor_planes = args.motion_feature_dim
        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 128 - 2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class BasicFuseMotion(nn.Module):
    def __init__(self, args):
        super(BasicFuseMotion, self).__init__()
        cor_planes = args.motion_feature_dim
        out_planes = args.query_latent_dim

        self.normf1 = nn.InstanceNorm2d(128)
        self.normf2 = nn.InstanceNorm2d(128)

        self.convf1 = nn.Conv2d(2, 128, 3, padding=1)
        self.convf2 = nn.Conv2d(128, 128, 3, padding=1)
        self.convf3 = nn.Conv2d(128, 64, 3, padding=1)

        s = 1
        self.normc1 = nn.InstanceNorm2d(256 * s)
        self.normc2 = nn.InstanceNorm2d(256 * s)
        self.normc3 = nn.InstanceNorm2d(256 * s)

        self.convc1 = nn.Conv2d(cor_planes + 128, 256 * s, 1, padding=0)
        self.convc2 = nn.Conv2d(256 * s, 256 * s, 3, padding=1)
        self.convc3 = nn.Conv2d(256 * s, 256 * s, 3, padding=1)
        self.convc4 = nn.Conv2d(256 * s, 256 * s, 3, padding=1)
        self.conv = nn.Conv2d(256 * s + 64, out_planes, 1, padding=0)

    def forward(self, flow, feat, context1=None):
        flo = F.relu(self.normf1(self.convf1(flow)))
        flo = F.relu(self.normf2(self.convf2(flo)))
        flo = self.convf3(flo)

        feat = torch.cat([feat, context1], dim=1)
        feat = F.relu(self.normc1(self.convc1(feat)))
        feat = F.relu(self.normc2(self.convc2(feat)))
        feat = F.relu(self.normc3(self.convc3(feat)))
        feat = self.convc4(feat)

        feat = torch.cat([flo, feat], dim=1)
        feat = F.relu(self.conv(feat))

        return feat


class BasicUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128, input_dim=128):
        super(BasicUpdateBlock, self).__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class DirectMeanMaskPredictor(nn.Module):
    def __init__(self, args):
        super(DirectMeanMaskPredictor, self).__init__()
        self.flow_head = FlowHead(args.predictor_dim, hidden_dim=256)
        self.mask = nn.Sequential(
            nn.Conv2d(args.predictor_dim, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, motion_features):
        delta_flow = self.flow_head(motion_features)
        mask = 0.25 * self.mask(motion_features)

        return mask, delta_flow


class BaiscMeanPredictor(nn.Module):
    def __init__(self, args, hidden_dim=128):
        super(BaiscMeanPredictor, self).__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, latent, flow):
        motion_features = self.encoder(flow, latent)
        delta_flow = self.flow_head(motion_features)
        mask = 0.25 * self.mask(motion_features)

        return mask, delta_flow


class BasicRPEEncoder(nn.Module):
    def __init__(self, args):
        super(BasicRPEEncoder, self).__init__()
        self.args = args
        dim = args.query_latent_dim
        self.encoder = nn.Sequential(
            nn.Linear(2, dim // 2),
            nn.ReLU(inplace=True),
            nn.Linear(dim // 2, dim),
            nn.ReLU(inplace=True),
            nn.Linear(dim, dim),
        )

    def forward(self, rpe_tokens):
        return self.encoder(rpe_tokens)


from .twins import Block, CrossBlock


class TwinsSelfAttentionLayer(nn.Module):
    def __init__(self, args):
        super(TwinsSelfAttentionLayer, self).__init__()
        self.args = args
        embed_dim = 256
        num_heads = 8
        mlp_ratio = 4
        ws = 7
        sr_ratio = 4
        dpr = 0.0
        drop_rate = 0.0
        attn_drop_rate = 0.0

        self.local_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=ws,
            with_rpe=True,
        )
        self.global_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=1,
            with_rpe=True,
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1.0)
            m.bias.data.zero_()

    def forward(self, x, tgt, size):
        x = self.local_block(x, size)
        x = self.global_block(x, size)

        tgt = self.local_block(tgt, size)
        tgt = self.global_block(tgt, size)
        return x, tgt


class TwinsCrossAttentionLayer(nn.Module):
    def __init__(self, args):
        super(TwinsCrossAttentionLayer, self).__init__()
        self.args = args
        embed_dim = 256
        num_heads = 8
        mlp_ratio = 4
        ws = 7
        sr_ratio = 4
        dpr = 0.0
        drop_rate = 0.0
        attn_drop_rate = 0.0

        self.local_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=ws,
            with_rpe=True,
        )
        self.global_block = CrossBlock(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=1,
            with_rpe=True,
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1.0)
            m.bias.data.zero_()

    def forward(self, x, tgt, size):
        x = self.local_block(x, size)
        tgt = self.local_block(tgt, size)
        x, tgt = self.global_block(x, tgt, size)

        return x, tgt



# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.

"""
Object detection and instance segmentation with XCiT backbone
Based on mmseg, timm and DeiT code bases
https://github.com/open-mmlab/mmsegmentation
https://github.com/rwightman/pytorch-image-models/tree/master/timm
https://github.com/facebookresearch/deit/
"""
import math
from functools import partial
import torch
import torch.nn as nn
from timm.models.vision_transformer import Mlp
from timm.models.layers import DropPath, trunc_normal_, to_2tuple

""" MLP module w/ dropout and configurable activation layer
Hacked together by / Copyright 2020 Ross Wightman
"""
from .helpers import to_2tuple


class Mlp(nn.Module):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        bias=True,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)

        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)
        return x


class PositionalEncodingFourier(nn.Module):
    """
    Positional encoding relying on a fourier kernel matching the one used in the
    "Attention is all of Need" paper. The implementation builds on DeTR code
    https://github.com/facebookresearch/detr/blob/master/models/position_encoding.py
    """

    def __init__(self, hidden_dim=32, dim=768, temperature=10000):
        super().__init__()
        self.token_projection = nn.Conv2d(hidden_dim * 2, dim, kernel_size=1)
        self.scale = 2 * math.pi
        self.temperature = temperature
        self.hidden_dim = hidden_dim
        self.dim = dim

    def forward(self, B, H, W, dtype):
        mask = torch.zeros(B, H, W).bool().to(self.token_projection.weight.device)
        not_mask = ~mask
        y_embed = not_mask.cumsum(1, dtype=dtype)
        x_embed = not_mask.cumsum(2, dtype=dtype)
        eps = 1e-6
        y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
        x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.hidden_dim, dtype=dtype, device=mask.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.hidden_dim)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack(
            (pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos_y = torch.stack(
            (pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        pos = self.token_projection(pos)
        return pos


class LPI(nn.Module):
    """
    Local Patch Interaction module that allows explicit communication between tokens in 3x3 windows
    to augment the implicit communcation performed by the block diagonal scatter attention.
    Implemented using 2 layers of separable 3x3 convolutions with GeLU and BatchNorm2d
    """

    def __init__(
        self, in_features, out_features=None, act_layer=nn.GELU, kernel_size=3
    ):
        super().__init__()
        out_features = out_features or in_features

        padding = kernel_size // 2

        self.conv1 = torch.nn.Conv2d(
            in_features,
            out_features,
            kernel_size=kernel_size,
            padding=padding,
            groups=out_features,
        )
        self.act = act_layer()
        self.bn = nn.GroupNorm(num_groups=8, num_channels=in_features)
        self.conv2 = torch.nn.Conv2d(
            in_features,
            out_features,
            kernel_size=kernel_size,
            padding=padding,
            groups=out_features,
        )

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.permute(0, 2, 1).reshape(B, C, H, W)
        x = self.conv1(x)
        x = self.act(x)
        x = self.bn(x)
        x = self.conv2(x)
        x = x.reshape(B, C, N).permute(0, 2, 1)

        return x


class XCA(nn.Module):
    """Cross-Covariance Attention (XCA) operation where the channels are updated using a weighted
     sum. The weights are obtained from the (softmax normalized) Cross-covariance
    matrix (Q^T K \\in d_h \\times d_h)
    """

    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        print("Self-att XCiT")

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)
        qkv = qkv.permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]
        q = q.transpose(-2, -1)
        k = k.transpose(-2, -1)
        v = v.transpose(-2, -1)

        q = torch.nn.functional.normalize(q, dim=-1)
        k = torch.nn.functional.normalize(k, dim=-1)

        attn = (q @ k.transpose(-2, -1)) * self.temperature
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        a = attn @ v
        x = (a).permute(0, 3, 1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    @torch.jit.ignore
    def no_weight_decay(self):
        return {"temperature"}


class XCA_separate(nn.Module):
    """Cross-Covariance Attention (XCA) operation where the channels are updated using a weighted
     sum. The weights are obtained from the (softmax normalized) Cross-covariance
    matrix (Q^T K \\in d_h \\times d_h)
    """

    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0.0, proj_drop=0.0):
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.to_qk = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.to_v = nn.Linear(dim, dim, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        print("Cross-att XCA")

    def forward(self, x_qk, x_v):
        B, N, C = x_qk.shape

        qk = self.to_qk(x_qk).reshape(B, N, 2, self.num_heads, C // self.num_heads)
        v = self.to_v(x_v).reshape(B, N, 1, self.num_heads, C // self.num_heads)
        qk = qk.permute(2, 0, 3, 1, 4)
        v = v.permute(2, 0, 3, 1, 4)
        q, k = qk[0], qk[1]

        q = q.transpose(-2, -1)
        k = k.transpose(-2, -1)
        v = v.transpose(-2, -1)

        q = torch.nn.functional.normalize(q, dim=-1)
        k = torch.nn.functional.normalize(k, dim=-1)

        attn = (q @ k.transpose(-2, -1)) * self.temperature
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ torch.squeeze(v, dim=0)).permute(0, 3, 1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    @torch.jit.ignore
    def no_weight_decay(self):
        return {"temperature"}


class XCABlock(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        mlp_ratio=1,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        separate=False,
        eta=None,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.separate = separate
        if separate:
            self.attn = XCA_separate(
                dim, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop
            )
        else:
            self.attn = XCA(
                dim,
                num_heads=num_heads,
                qkv_bias=qkv_bias,
                attn_drop=attn_drop,
                proj_drop=drop,
            )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)

        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

        self.norm3 = norm_layer(dim)
        self.local_mp = LPI(in_features=dim, act_layer=act_layer)

        self.gamma1 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
        self.gamma2 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)
        self.gamma3 = nn.Parameter(eta * torch.ones(dim), requires_grad=True)

    def forward(self, x, H, W, x_v=None):
        if self.separate:
            x = x + self.drop_path(
                self.gamma1 * self.attn(x_qk=self.norm1(x), x_v=self.norm1(x_v))
            )
        else:
            x = x + self.drop_path(self.gamma1 * self.attn(self.norm1(x)))
        x = x + self.drop_path(self.gamma3 * self.local_mp(self.norm3(x), H, W))
        x = x + self.drop_path(self.gamma2 * self.mlp(self.norm2(x)))
        return x


# @BACKBONES.register_module() # I commented this
class XCiT(nn.Module):
    """
    Based on timm and DeiT code bases
    https://github.com/rwightman/pytorch-image-models/tree/master/timm
    https://github.com/facebookresearch/deit/
    """

    def __init__(
        self,
        embed_dim=128,
        depth=1,
        num_heads=8,
        mlp_ratio=1,
        qkv_bias=True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.05,
        norm_layer=None,
        use_pos=True,
        eta=1.0,
        separate=False,
    ):
        """
        Args:
            img_size (int, tuple): input image size
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            qk_scale (float): override default qk scale of head_dim ** -0.5 if set
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            norm_layer: (nn.Module): normalization layer
            use_pos: (bool) whether to use positional encoding
            eta: (float) layerscale initialization value
        """
        super().__init__()
        self.num_features = self.embed_dim = embed_dim
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        self.separate = separate

        #  NO PATCHIFICATION OF XCiT's INPUTS-- Inputs used in their original resolution.
        # SKIPED PATCHIFICATION-related steps...

        dpr = [drop_path_rate for i in range(depth)]
        if self.separate:
            depth = 1
        self.blocks = nn.ModuleList(
            [
                XCABlock(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                    eta=eta,
                    separate=self.separate,
                )
                for i in range(depth)
            ]
        )

        self.pos_embeder = PositionalEncodingFourier(dim=embed_dim)
        self.use_pos = use_pos

        # SKIPED PATCHIFICATION-related steps...

    @torch.jit.ignore
    def no_weight_decay(self):
        return {"pos_embed", "cls_token", "dist_token"}

    def init_weights(self, pretrained=None):
        """Initialize the weights in backbone.
        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """

        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=0.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        if isinstance(pretrained, str):
            self.apply(_init_weights)
        elif pretrained is None:
            self.apply(_init_weights)
        else:
            raise TypeError("pretrained must be a str or None")

    # self- and cross- attention:
    def forward(self, x, x_v=None):
        B, C, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        pos_encoding = (
            self.pos_embeder(B, H, W, x.dtype)
            .reshape(B, -1, x.shape[1])
            .permute(0, 2, 1)
        )
        if self.separate:
            x_v = x_v.flatten(2).transpose(1, 2)
        if self.use_pos:
            x = x + pos_encoding

        for blk in self.blocks:
            if self.separate:
                x = blk(x, x_v=x_v, H=H, W=W)
            else:
                x = blk(x, H, W)

        x = x.permute(0, 2, 1).reshape(B, -1, H, W)
        # SKIPED PATCHIFICATION-related steps...
        return x



import torch.nn.functional as F


class InputPadder:
    def __init__(self, dims, base=8):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // base) + 1) * base - self.ht) % base
        pad_wd = (((self.wd // base) + 1) * base - self.wd) % base
        self._pad = [0, pad_wd, 0, pad_ht]

    def pad(self, *inputs):
        outputs = []

        for x in inputs:
            bhw_mode = 0

            if len(x.shape) == 3:
                bhw_mode = 1
                x = x.unsqueeze(1)

            x = F.pad(x, self._pad, mode="replicate")
            if bhw_mode:
                x = x.squeeze(1)
            outputs.append(x)

        return outputs

    def unpad(self, *inputs):
        ht, wd = inputs[0].shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]

        return [x[..., c[0] : c[1], c[2] : c[3]] for x in inputs]



import torch
import torch.nn.functional as F


class ConvBlock(torch.nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding):
        super(ConvBlock, self).__init__()

        self.conv1 = torch.nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode="zeros",
            bias=False,
        )

        self.conv2 = torch.nn.Conv2d(
            out_planes, out_planes, kernel_size=3, stride=1, padding=1, bias=False
        )

        self.relu = torch.nn.LeakyReLU(negative_slope=0.1, inplace=False)

        self.norm1 = torch.nn.BatchNorm2d(out_planes)

        self.norm2 = torch.nn.BatchNorm2d(out_planes)

        # self.dropout = torch.nn.Dropout(p=0.1)

    def forward(self, x):
        # x = self.dropout(x)

        x = self.relu(self.norm1(self.conv1(x)))
        x = self.relu(self.norm2(self.conv2(x)))
        # x = self.relu(self.conv1(x))
        # x = self.relu(self.conv2(x))

        return x

    def forward_fuse(self, x):
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))

        return x


class CNNEncoder(torch.nn.Module):
    def __init__(
        self, feature_dim_s16, context_dim_s16, feature_dim_s8, context_dim_s8
    ):
        super(CNNEncoder, self).__init__()

        self.block_8_1 = ConvBlock(
            3, feature_dim_s8 * 2, kernel_size=8, stride=4, padding=2
        )

        self.block_8_2 = ConvBlock(
            3, feature_dim_s8, kernel_size=6, stride=2, padding=2
        )

        self.block_cat_8 = ConvBlock(
            feature_dim_s8 * 3,
            feature_dim_s8 + context_dim_s8,
            kernel_size=3,
            stride=1,
            padding=1,
        )

        self.block_16_1 = ConvBlock(
            3, feature_dim_s16, kernel_size=6, stride=2, padding=2
        )

        self.block_8_16 = ConvBlock(
            feature_dim_s8 + context_dim_s8,
            feature_dim_s16,
            kernel_size=6,
            stride=2,
            padding=2,
        )

        self.block_cat_16 = ConvBlock(
            feature_dim_s16 * 2,
            feature_dim_s16 + context_dim_s16 - 2,
            kernel_size=3,
            stride=1,
            padding=1,
        )

    def init_pos(self, batch_size, height, width, device, amp):
        ys, xs = torch.meshgrid(
            torch.arange(
                height, dtype=torch.half if amp else torch.float, device=device
            ),
            torch.arange(
                width, dtype=torch.half if amp else torch.float, device=device
            ),
            indexing="ij",
        )
        ys = ys - height / 2
        xs = xs - width / 2
        pos = torch.stack([ys, xs])
        return pos[None].repeat(batch_size, 1, 1, 1)

    def init_bhwd(self, batch_size, height, width, device, amp):
        self.pos_s16 = self.init_pos(batch_size, height, width, device, amp)

    def forward(self, img):
        img = F.avg_pool2d(img, kernel_size=2, stride=2)
        x_8 = self.block_8_1(img)

        img = F.avg_pool2d(img, kernel_size=2, stride=2)
        x_8_2 = self.block_8_2(img)

        x_8 = self.block_cat_8(torch.cat([x_8, x_8_2], dim=1))

        img = F.avg_pool2d(img, kernel_size=2, stride=2)
        x_16 = self.block_16_1(img)

        x_16_2 = self.block_8_16(x_8)

        x_16 = self.block_cat_16(torch.cat([x_16, x_16_2], dim=1))

        x_16 = torch.cat([x_16, self.pos_s16], dim=1)

        return x_16, x_8



import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import GMAUpdateBlock
from ..encoders import twins_svt_large
from .cnn import BasicEncoder
from .corr import CorrBlock
from ..utils import coords_grid
from .gma import Attention
from .sk import SKUpdateBlock6_Deep_nopoolres_AllDecoder
from .sk2 import SKUpdateBlock6_Deep_nopoolres_AllDecoder2_Mem_skflow
from .memory_util import *

try:
    from flash_attn import flash_attn_func
except:
    pass


class MemFlowNet(nn.Module):
    def __init__(
        self,
        cnet,
        fnet,
        corr_levels,
        corr_radius,
        decoder_depth,
        pretrain,
        gma,
        corr_fn,
        train_avg_length,
    ):
        super().__init__()

        self.cnet_name = cnet
        self.fnet_name = fnet
        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.decoder_depth = decoder_depth

        self.hidden_dim = 128
        self.context_dim = 128

        # feature network, context network, and update block
        if cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=pretrain)
            self.proj = nn.Conv2d(256, 256, 1)
        elif cnet == "basicencoder":
            print("[Using basicencoder as context encoder]")
            self.cnet = BasicEncoder(output_dim=256, norm_fn="batch")

        if fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=pretrain)
            self.channel_convertor = nn.Conv2d(256, 256, 1, padding=0, bias=False)
        elif fnet == "basicencoder":
            print("[Using basicencoder as feature encoder]")
            self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if gma == "GMA":
            print("[Using GMA]")
            self.update_block = GMAUpdateBlock(
                corr_levels=corr_levels, corr_radius=corr_radius, hidden_dim=128
            )
        elif gma == "GMA-SK":
            print("[Using GMA-SK]")
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder(
                cost_heads_num=1, hidden_dim=128
            )
        elif gma == "GMA-SK2":
            print("[Using GMA-SK2]")
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2_Mem_skflow(
                corr_radius=corr_radius,
                corr_levels=corr_levels,
                cost_heads_num=1,
                hidden_dim=128,
            )

        print("[Using corr_fn {}]".format(corr_fn))

        self.att = Attention(
            dim=self.context_dim,
            heads=1,
            max_pos_size=160,
            dim_head=self.context_dim,
        )
        self.train_avg_length = train_avg_length

    def encode_features(self, frame, flow_init=None):
        # Determine input shape
        if len(frame.shape) == 5:
            # shape is b*t*c*h*w
            need_reshape = True
            b, t = frame.shape[:2]
            # flatten so that we can feed them into a 2D CNN
            frame = frame.flatten(start_dim=0, end_dim=1)
        elif len(frame.shape) == 4:
            # shape is b*c*h*w
            need_reshape = False
        else:
            raise NotImplementedError

        fmaps = self.fnet(frame)
        if self.fnet_name == "twins":
            fmaps = self.channel_convertor(fmaps)
        if need_reshape:
            # B*T*C*H*W
            fmaps = fmaps.view(b, t, *fmaps.shape[-3:])
            frame = frame.view(b, t, *frame.shape[-3:])
            coords0, coords1 = self.initialize_flow(frame[:, 0, ...])
        else:
            coords0, coords1 = self.initialize_flow(frame)
        if flow_init is not None:
            coords1 = coords1 + flow_init

        return coords0, coords1, fmaps

    def encode_context(self, frame):
        # Determine input shape
        if len(frame.shape) == 5:
            # shape is b*t*c*h*w
            need_reshape = True
            b, t = frame.shape[:2]
            # flatten so that we can feed them into a 2D CNN
            frame = frame.flatten(start_dim=0, end_dim=1)
        elif len(frame.shape) == 4:
            # shape is b*c*h*w
            need_reshape = False
        else:
            raise NotImplementedError

        # shape is b*c*h*w
        cnet = self.cnet(frame)
        if self.cnet_name == "twins":
            cnet = self.proj(cnet)

        net, inp = torch.split(cnet, [self.hidden_dim, self.context_dim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        query, key = self.att.to_qk(inp).chunk(2, dim=1)
        # query = query * self.att.scale
        if need_reshape:
            # B*C*T*H*W
            query = query.view(b, t, *query.shape[-3:]).transpose(1, 2).contiguous()
            key = key.view(b, t, *key.shape[-3:]).transpose(1, 2).contiguous()

            # B*T*C*H*W
            net = net.view(b, t, *net.shape[-3:])
            inp = inp.view(b, t, *inp.shape[-3:])

        return query, key, net, inp

    def predict_flow(
        self,
        net,
        inp,
        coords0,
        coords1,
        fmaps,
        query,
        ref_keys,
        ref_values,
        test_mode=False,
    ):
        corr_fn = CorrBlock(
            fmaps[:, 0, ...],
            fmaps[:, 1, ...],
            num_levels=self.corr_levels,
            radius=self.corr_radius,
        )
        flow_predictions = []
        query = query.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
        ref_keys = ref_keys.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
        for _ in range(self.decoder_depth):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume
            flow = coords1 - coords0
            motion_features, current_value = self.update_block.get_motion_and_value(
                flow, corr
            )
            current_value = current_value.unsqueeze(2)
            value = (
                current_value
                if ref_values is None
                else torch.cat([ref_values, current_value], dim=2)
            )
            # get global motion
            # B, L, N, C
            value = value.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
            scale = self.att.scale * math.log(ref_keys.shape[1], self.train_avg_length)
            hidden_states = flash_attn_func(
                query, ref_keys, value, dropout_p=0.0, softmax_scale=scale, causal=False
            )
            hidden_states = (
                hidden_states.squeeze(2).permute(0, 2, 1).reshape(motion_features.shape)
            )

            motion_features_global = (
                motion_features + self.update_block.aggregator.gamma * hidden_states
            )
            net, up_mask, delta_flow = self.update_block(
                net, inp, motion_features, motion_features_global
            )
            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow
            # upsample predictions
            flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_predictions.append(flow_up)

        if test_mode:
            return coords1 - coords0, flow_up, current_value
        else:
            return flow_predictions, current_value

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, device=img.device, dtype=img.dtype)
        coords1 = coords_grid(N, H // 8, W // 8, device=img.device, dtype=img.dtype)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)



import torch
import torch.nn as nn
import copy


class Attention1D(nn.Module):
    """Cross-Attention on x or y direction,
    without multi-head and dropout support for faster speed
    """

    def __init__(
        self,
        in_channels,
        y_attention=False,
        double_cross_attn=False,  # cross attn feature1 before computing cross attn feature2
        **kwargs,
    ):
        super(Attention1D, self).__init__()

        self.y_attention = y_attention
        self.double_cross_attn = double_cross_attn

        # self attn feature1 before cross attn
        if double_cross_attn:
            self.self_attn = copy.deepcopy(
                Attention1D(
                    in_channels=in_channels,
                    y_attention=not y_attention,
                )
            )

        self.query_conv = nn.Conv2d(in_channels, in_channels, 1)
        self.key_conv = nn.Conv2d(in_channels, in_channels, 1)

        # Initialize: https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/transformer.py#L138
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)  # original Transformer initialization

    def forward(self, feature1, feature2, position=None, value=None):
        b, c, h, w = feature1.size()

        # self attn before cross attn
        if self.double_cross_attn:
            feature1 = self.self_attn(feature1, feature1, position)[
                0
            ]  # self attn feature1

        query = feature1 + position if position is not None else feature1
        query = self.query_conv(query)  # [B, C, H, W]

        key = feature2 + position if position is not None else feature2

        key = self.key_conv(key)  # [B, C, H, W]
        value = feature2 if value is None else value  # [B, C, H, W]
        scale_factor = c**0.5

        if self.y_attention:
            query = query.permute(0, 3, 2, 1)  # [B, W, H, C]
            key = key.permute(0, 3, 1, 2)  # [B, W, C, H]
            value = value.permute(0, 3, 2, 1)  # [B, W, H, C]
        else:  # x attention
            query = query.permute(0, 2, 3, 1)  # [B, H, W, C]
            key = key.permute(0, 2, 1, 3)  # [B, H, C, W]
            value = value.permute(0, 2, 3, 1)  # [B, H, W, C]

        scores = torch.matmul(query, key) / scale_factor  # [B, W, H, H] or [B, H, W, W]

        attention = torch.softmax(scores, dim=-1)  # [B, W, H, H] or [B, H, W, W]

        out = torch.matmul(attention, value)  # [B, W, H, C] or [B, H, W, C]

        if self.y_attention:
            out = out.permute(0, 3, 2, 1).contiguous()  # [B, C, H, W]
        else:
            out = out.permute(0, 3, 1, 2).contiguous()  # [B, C, H, W]

        return out, attention



"""Provide useful functions for using PTLFlow."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__version__ = "0.4.1"

from pathlib import Path
from typing import Any, Dict, List, Optional

from jsonargparse import ArgumentParser, Namespace
from loguru import logger
import requests
import torch
from torch import hub

from ptlflow.models.base_model.base_model import BaseModel

from ptlflow.utils.registry import (
    _models_dict,
    _ptlflow_trained_models,
    _trainable_models,
)


def download_scripts(destination_dir: Path = Path("ptlflow_scripts")) -> None:
    """Download the main scripts and configs to start working with PTLFlow."""
    github_url = "https://raw.githubusercontent.com/hmorimitsu/ptlflow/main/"
    script_names = [
        "datasets.yaml",
        "infer.py",
        "model_benchmark.py",
        "test.py",
        "train.py",
        "validate.py",
    ]

    destination_dir.mkdir(parents=True, exist_ok=True)

    for sname in script_names:
        script_url = github_url + sname
        data = requests.get(script_url)
        if data.status_code == 200:
            with open(destination_dir / sname, "wb") as f:
                f.write(data.content)
        else:
            logger.warning("Script {} was not found.", script_url)

    logger.info("Downloaded scripts to {}.", str(destination_dir))


def get_model(
    model_name: str,
    ckpt_path: Optional[str] = None,
    args: Optional[Namespace] = None,
) -> BaseModel:
    """Return an instance of a chosen model.

    The instance can have configured by he arguments, and load some existing pretrained weights.

    Note that this is different from get_model_reference(), which returns a reference to the model class. The instance,
    returned by this function, is a class already instantiated. Therefore, the return of this function is equivalent to
    "return get_model_reference()()", which looks confusing. This can be rewritten as
    "model_ref = get_model_reference(); return model_ref()".

    Parameters
    ----------
    model_name : str
        Name of the model to get an instance of.
    ckpt_path : Optional[str], optional
        Name of the pretrained weight to load or a path to a local checkpoint file.
    args : Optional[Namespace], optional
        Some arguments that ill be provided to the model.

    Returns
    -------
    BaseModel
        The instance of the chosen model.

    Raises
    ------
    ValueError
        If the given checkpoint name is not a valid choice.
    ValueError
        If a checkpoint name is given, but the model does not have any pretrained weights available.

    See Also
    --------
    get_model_reference : To get a reference to the class of a model.
    """
    model_ref = get_model_reference(model_name)
    if args is None:
        parser = ArgumentParser()
        parser.add_class_arguments(model_ref, "model")
        args = parser.parse_args([])

    model_parser = ArgumentParser(exit_on_error=False)
    model_parser.add_argument("--model", type=model_ref)
    model_cfg = model_parser.parse_object({"model": args.model})
    model = model_parser.instantiate_classes(model_cfg).model

    if (
        ckpt_path is None
        and args is not None
        and hasattr(args, "ckpt_path")
        and args.ckpt_path is not None
    ):
        ckpt_path = args.ckpt_path

    model = restore_model(model, ckpt_path)

    return model


def get_model_reference(model_name: str) -> BaseModel:
    """Return a reference to the class of a chosen model.

    Note that this is different from get_model(), which returns an instance of a model. The reference, returned by this
    function, is a class before instantiation. Therefore, the return of this function can be used to instantiate a model as
    "model_ref = get_model_reference(); model_instance = model_ref()".

    Parameters
    ----------
    model_name : str
        Name of the model to get a reference of.

    Returns
    -------
    BaseModel
        A reference to the chosen model.

    Raises
    ------
    ValueError
        If the given name is not a valid choice.

    See Also
    --------
    get_model : To get an instance of a model.
    """
    try:
        return _models_dict[model_name]
    except KeyError:
        raise ValueError(
            f'Unknown model name: {model_name}. Choose from [{", ".join(_models_dict.keys())}]'
        )


def get_model_names() -> List[str]:
    """Return a list of all model names that are registered in this platform.

    Models are added to this list by decorating their classes with @ptlflow.utils.registry.register_model.

    Returns
    -------
    List[str]
        The list of the all registered model names.
    """
    return sorted(list(_models_dict.keys()))


def get_trainable_model_names() -> List[str]:
    """Return a list of model names that are able to be trained.

    This function return the names of the model that have a loss function defined.

    Returns
    -------
    List[str]
        The list of the model names that can be trained.
    """
    return _trainable_models


def get_ptlflow_trained_model_names() -> List[str]:
    """Return a list of model names that have been trained on PTLFlow.

    This function return the names of the model that have set model.has_trained_on_ptlflow = True.

    Returns
    -------
    List[str]
        The list of the model names that has been trained on PTLFlow.
    """
    return _ptlflow_trained_models


def load_checkpoint(ckpt_path: str, model_ref: BaseModel) -> Dict[str, Any]:
    """Try to load the checkpoint specified in ckpt_path.

    Parameters
    ----------
    ckpt_path : str
        Path to a local file or name of a pretrained checkpoint.
    model_ref : BaseModel
        A reference to the model class. See the function get_model_reference() for more details.

    Returns
    -------
    Dict[str, Any]
        A dictionary of the loaded checkpoint. The output of torch.load().

    See Also
    --------
    get_model_reference : To get a reference to the class of a model.
    """
    if Path(ckpt_path).exists():
        ckpt_path = ckpt_path
    elif hasattr(model_ref, "pretrained_checkpoints"):
        tmp_ckpt_path = model_ref.pretrained_checkpoints.get(ckpt_path)
        if tmp_ckpt_path is None:
            raise ValueError(
                f"Invalid checkpoint name {ckpt_path}. "
                f'Choose one from {{{",".join(model_ref.pretrained_checkpoints.keys())}}}'
            )
        else:
            ckpt_path = tmp_ckpt_path
    else:
        raise ValueError(
            f"Cannot find checkpoint {ckpt_path} for model {model_ref.__name__}"
        )

    device = "cuda" if torch.cuda.is_available() else "cpu"

    if Path(ckpt_path).exists():
        ckpt = torch.load(
            ckpt_path, map_location=torch.device(device), weights_only=True
        )
    else:
        model_dir = Path(hub.get_dir()) / "checkpoints"
        ckpt = hub.load_state_dict_from_url(
            ckpt_path,
            model_dir=model_dir,
            map_location=torch.device(device),
            check_hash=True,
            weights_only=True,
        )
    return ckpt


def restore_model(model, ckpt_path):
    """Load model state from the checkpoint.

    Parameters
    ----------
    model : BaseModel
        An instance of the model to be restored.
    ckpt_path : str
        Path to a local file or name of a pretrained checkpoint.

    Returns
    -------
    BaseModel
        An instance of the restored model.
    """
    if ckpt_path is not None:
        ckpt = load_checkpoint(ckpt_path, model.__class__)

        state_dict = ckpt["state_dict"]
        if "hyper_parameters" in ckpt:
            if "train_size" in ckpt["hyper_parameters"]:
                model.train_size = ckpt["hyper_parameters"]["train_size"]
            if "train_avg_length" in ckpt["hyper_parameters"]:
                model.train_avg_length = ckpt["hyper_parameters"]["train_avg_length"]
            if "extra_params" in ckpt["hyper_parameters"]:
                extra_params = ckpt["hyper_parameters"]["extra_params"]
                for name, value in extra_params.items():
                    model.add_extra_param(name, value)
        model.load_state_dict(state_dict)
        logger.info("Restored model state from checkpoint: {}", ckpt_path)

    return model



import torch
from torch.optim.lr_scheduler import (
    MultiStepLR,
    CosineAnnealingLR,
    ExponentialLR,
    OneCycleLR,
)


def fetch_optimizer(model, cfg):
    """Create the optimizer and learning rate scheduler"""
    # optimizer = optim.AdamW(model.parameters(), lr=args.lr, weight_decay=args.wdecay, eps=args.epsilon)

    # scheduler = optim.lr_scheduler.OneCycleLR(optimizer, args.lr, args.num_steps+100,
    #     pct_start=0.05, cycle_momentum=False, anneal_strategy='linear')
    optimizer = build_optimizer(model, cfg)
    scheduler = build_scheduler(cfg, optimizer)

    return optimizer, scheduler


def build_optimizer(model, config):
    name = config.optimizer
    lr = config.canonical_lr

    if name == "adam":
        return torch.optim.Adam(
            model.parameters(),
            lr=lr,
            weight_decay=config.adam_decay,
            eps=config.epsilon,
        )
    elif name == "adamw":
        if hasattr(config, "twins_lr_factor"):
            factor = config.twins_lr_factor
            print("[Decrease lr of pre-trained model by factor {}]".format(factor))
            param_dicts = [
                {
                    "params": [
                        p
                        for n, p in model.named_parameters()
                        if "feat_encoder" not in n
                        and "context_encoder" not in n
                        and p.requires_grad
                    ]
                },
                {
                    "params": [
                        p
                        for n, p in model.named_parameters()
                        if ("feat_encoder" in n or "context_encoder" in n)
                        and p.requires_grad
                    ],
                    "lr": lr * factor,
                },
            ]
            full = [n for n, _ in model.named_parameters()]
            return torch.optim.AdamW(
                param_dicts, lr=lr, weight_decay=config.adamw_decay, eps=config.epsilon
            )
        else:
            return torch.optim.AdamW(
                model.parameters(),
                lr=lr,
                weight_decay=config.adamw_decay,
                eps=config.epsilon,
            )
    else:
        raise ValueError(f"TRAINER.OPTIMIZER = {name} is not a valid optimizer!")


def build_scheduler(config, optimizer):
    """
    Returns:
        scheduler (dict):{
            'scheduler': lr_scheduler,
            'interval': 'step',  # or 'epoch'
        }
    """
    # scheduler = {'interval': config.TRAINER.SCHEDULER_INTERVAL}
    name = config.scheduler
    lr = config.canonical_lr

    if name == "OneCycleLR":
        # scheduler = OneCycleLR(optimizer, )
        if hasattr(config, "twins_lr_factor"):
            factor = config.twins_lr_factor
            scheduler = OneCycleLR(
                optimizer,
                [lr, lr * factor],
                config.num_steps + 100,
                pct_start=0.05,
                cycle_momentum=False,
                anneal_strategy=config.anneal_strategy,
            )
        else:
            scheduler = OneCycleLR(
                optimizer,
                lr,
                config.num_steps + 100,
                pct_start=0.05,
                cycle_momentum=False,
                anneal_strategy=config.anneal_strategy,
            )
    # elif name == 'MultiStepLR':
    #     scheduler.update(
    #         {'scheduler': MultiStepLR(optimizer, config.TRAINER.MSLR_MILESTONES, gamma=config.TRAINER.MSLR_GAMMA)})
    # elif name == 'CosineAnnealing':
    #    scheduler = CosineAnnealingLR(optimizer, config.num_steps+100)
    #     scheduler.update(
    #         {'scheduler': CosineAnnealingLR(optimizer, config.TRAINER.COSA_TMAX)})
    # elif name == 'ExponentialLR':
    #     scheduler.update(
    #         {'scheduler': ExponentialLR(optimizer, config.TRAINER.ELR_GAMMA)})
    else:
        raise NotImplementedError()

    return scheduler



# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
import shutil

from jsonargparse import ArgumentParser
import ptlflow
import model_benchmark

TEST_MODEL = "raft_small"


def test_benchmark(tmp_path: Path) -> None:
    model_ref = ptlflow.get_model_reference(TEST_MODEL)

    model_parser = ArgumentParser(parents=[model_benchmark._init_parser()])
    model_parser.add_argument_group("model")
    model_parser.add_class_arguments(model_ref, "model.init_args")
    args = model_parser.parse_args([])
    args.model.class_path = f"{model_ref.__module__}.{model_ref.__qualname__}"

    args.num_samples = 1
    args.output_path = tmp_path

    model_benchmark.benchmark(args, None)

    shutil.rmtree(tmp_path)



# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
import shutil

import summary_metrics


def test_summary(tmp_path: Path) -> None:
    parser = summary_metrics._init_parser()
    args = parser.parse_args([])
    args.metrics_path = Path("docs/source/results/metrics_all.csv")
    args.output_dir = tmp_path
    summary_metrics.summarize(args)

    assert len(list(tmp_path.glob("**/*.md"))) > 0
    assert len(list(tmp_path.glob("**/*.csv"))) > 0

    shutil.rmtree(tmp_path)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
from pathlib import Path
import re
from typing import Optional

import torch
from torch import hub
from fsspec.core import url_to_fs

import lightning.pytorch as pl
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch.trainer.connectors.checkpoint_connector import (
    _CheckpointConnector,
)
from lightning.pytorch.trainer.states import TrainerFn
from lightning.pytorch.utilities.migration import pl_legacy_patch
from lightning.pytorch.utilities.migration.utils import _pl_migrate_checkpoint
from lightning.pytorch.utilities.rank_zero import rank_zero_info

log = logging.getLogger(__name__)


class _PTLFlowCheckpointConnector(_CheckpointConnector):
    def __init__(self, trainer: pl.Trainer) -> None:
        super().__init__(trainer)

    def resume_start(
        self,
        checkpoint_path: Optional[_PATH] = None,
        model: Optional[pl.LightningModule] = None,
    ) -> None:
        """Attempts to pre-load the checkpoint file to memory, with the source path determined in this priority:

        1. from HPC weights if `checkpoint_path` is ``None`` and on SLURM or passed keyword `"hpc"`.
        2. from fault-tolerant auto-saved checkpoint if found
        3. from `checkpoint_path` file if provided
        4. don't restore

        """
        self._ckpt_path = checkpoint_path
        if not checkpoint_path:
            log.debug("`checkpoint_path` not specified. Skipping checkpoint loading.")
            return

        if not Path(checkpoint_path).exists():
            if model is not None:
                model_ref = model.__class__
                if hasattr(model_ref, "pretrained_checkpoints"):
                    checkpoint_path = model_ref.pretrained_checkpoints.get(
                        checkpoint_path
                    )
                    if checkpoint_path is None:
                        raise ValueError(
                            f"Invalid checkpoint name {checkpoint_path}. "
                            f'Choose one from {{{",".join(model_ref.pretrained_checkpoints.keys())}}}'
                        )

                    cache_path = (
                        Path(hub.get_dir())
                        / "checkpoints"
                        / checkpoint_path.split("/")[-1]
                    )
                    if cache_path.exists():
                        checkpoint_path = cache_path
            else:
                raise ValueError(
                    f"Cannot find checkpoint {checkpoint_path} for model {model.__class__.__name__}"
                )

        rank_zero_info(
            f"Restoring states from the checkpoint path at {checkpoint_path}"
        )
        with pl_legacy_patch():
            loaded_checkpoint = self.trainer.strategy.load_checkpoint(checkpoint_path)
        if not "pytorch-lightning_version" in loaded_checkpoint:
            loaded_checkpoint["pytorch-lightning_version"] = "1.9.5"
        self._loaded_checkpoint = _pl_migrate_checkpoint(
            loaded_checkpoint, checkpoint_path
        )

    def resume_end(self) -> None:
        """Signal the connector that all states have resumed and memory for the checkpoint object can be released."""
        assert self.trainer.state.fn is not None
        if self._ckpt_path:
            message = (
                "Restored all states"
                if self.trainer.state.fn == TrainerFn.FITTING
                else "Loaded model weights"
            )
            rank_zero_info(f"{message} from the checkpoint at {self._ckpt_path}")

        # free memory
        self._loaded_checkpoint = {}
        torch.cuda.empty_cache()

        # wait for all to catch up
        self.trainer.strategy.barrier("_PTLFlowCheckpointConnector.resume_end")

    def restore_training_state(self) -> None:
        """Restore the trainer state from the pre-loaded checkpoint.

        This includes the precision settings, loop progress, optimizer states and learning rate scheduler states.

        Modifications by Henrique Morimitsu:
        - restore_optimizers_and_schedulers before other states to raise an error earlier in case the ckpt does not have training states
        """
        if not self._loaded_checkpoint:
            return

        assert self.trainer.state.fn is not None
        if self.trainer.state.fn == TrainerFn.FITTING:
            # restore optimizers and schedulers state
            self.restore_optimizers_and_schedulers()

        # restore precision plugin (scaler etc.)
        self.restore_precision_plugin_state()

        # restore loops and their progress
        self.restore_loops()

    def _restore_modules_and_callbacks(
        self,
        checkpoint_path: Optional[_PATH] = None,
        model: Optional[pl.LightningModule] = None,
    ) -> None:
        # restore modules after setup
        self.resume_start(checkpoint_path, model)
        self.restore_model()
        self.restore_datamodule()
        if self.trainer.state.fn == TrainerFn.FITTING:
            # restore callback states
            self.restore_callbacks()

    @staticmethod
    def __max_ckpt_version_in_folder(
        dir_path: _PATH, name_key: str = "ckpt_"
    ) -> Optional[int]:
        """List up files in `dir_path` with `name_key`, then yield maximum suffix number.

        Args:
            dir_path: path of directory which may contain files whose name include `name_key`
            name_key: file name prefix
        Returns:
            None if no-corresponding-file else maximum suffix number

        """
        # check directory existence
        fs, uri = url_to_fs(str(dir_path))
        if not fs.exists(dir_path):
            return None

        # check corresponding file existence
        files = [os.path.basename(f["name"]) for f in fs.listdir(uri)]
        files = [x for x in files if name_key in x]
        if len(files) == 0:
            return None

        # extract suffix number
        ckpt_vs = []
        for name in files:
            name = name.split(name_key)[-1]
            name = re.sub("[^0-9]", "", name)
            ckpt_vs.append(int(name))

        return max(ckpt_vs)

    @staticmethod
    def __get_max_ckpt_path_from_folder(folder_path: _PATH) -> str:
        """Get path of maximum-epoch checkpoint in the folder."""
        max_suffix = _PTLFlowCheckpointConnector.__max_ckpt_version_in_folder(
            folder_path
        )
        ckpt_number = max_suffix if max_suffix is not None else 0
        return f"{folder_path}/hpc_ckpt_{ckpt_number}.ckpt"

    @staticmethod
    def hpc_save_path(folderpath: _PATH) -> str:
        max_suffix = _PTLFlowCheckpointConnector.__max_ckpt_version_in_folder(
            folderpath
        )
        ckpt_number = (max_suffix if max_suffix is not None else 0) + 1
        return os.path.join(folderpath, f"hpc_ckpt_{ckpt_number}.ckpt")



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
import torch.nn.functional as F


def upsample2d_as(inputs, target_as, mode="bilinear"):
    _, _, h, w = target_as.size()
    if inputs.shape[-2] != h or inputs.shape[-1] != w:
        inputs = F.interpolate(inputs, [h, w], mode=mode, align_corners=True)
    return inputs


def rescale_flow(flow, width_im, height_im, to_local=True):
    if to_local:
        u_scale = float(flow.size(3) / width_im)
        v_scale = float(flow.size(2) / height_im)
    else:
        u_scale = float(width_im / flow.size(3))
        v_scale = float(height_im / flow.size(2))

    u, v = flow.chunk(2, dim=1)
    u = u * u_scale
    v = v * v_scale

    return torch.cat([u, v], dim=1)



"""

Create a file with a list of samples names from the AutoFlow [1] dataset to be used as validation samples.

[1] Sun, Deqing et al. “AutoFlow: Learning a Better Training Set for Optical Flow.” CVPR. 2021.

"""

# =============================================================================
# Copyright 2022 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


from argparse import ArgumentParser, Namespace
import os
from pathlib import Path
import random

random.seed(42)

THIS_DIR = Path(os.path.abspath(os.path.dirname(__file__)))


def _init_parser() -> ArgumentParser:
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument("--autoflow_root", type=str, required=True)
    parser.add_argument(
        "--output_file", type=str, default=str(THIS_DIR / "AutoFlow_val.txt")
    )
    parser.add_argument("--val_percentage", type=float, default=0.05)
    return parser


def main(args: Namespace) -> None:
    """Run the split process.

    Parameters
    ----------
    args : argparse.Namespace
        Arguments for configuring the splitting.
    """
    parts_dirs = [f"static_40k_png_{i+1}_of_4" for i in range(4)]
    sample_dirs = []
    for pdir in parts_dirs:
        sample_dirs.extend(
            sorted(
                [
                    f.stem
                    for f in (Path(args.autoflow_root) / pdir).glob("*")
                    if f.is_dir()
                ]
            )
        )
    sample_dirs.sort()
    assert (
        len(sample_dirs) == 40000
    ), f"ERROR: AutoFlow dataset should have 40k samples, but found {len(sample_dirs)}."
    samples_per_table = {}
    for sdir in sample_dirs:
        table_idx = sdir.split("_")[1]
        if table_idx not in samples_per_table:
            samples_per_table[table_idx] = []
        samples_per_table[table_idx].append(sdir)
    assert (
        len(samples_per_table) == 300
    ), f"ERROR: AutoFlow dataset should have 300 tables, but found {len(samples_per_table)}."

    val_samples = []
    carryover_samples = 0.0
    for dir_list in samples_per_table.values():
        num_samples = len(dir_list)
        num_val_samples_float = args.val_percentage * num_samples + carryover_samples
        num_val_samples = int(num_val_samples_float)

        random.shuffle(dir_list)
        val_samples.extend(dir_list[:num_val_samples])

        carryover_samples = num_val_samples_float - num_val_samples

    val_samples.sort(key=lambda x: 1000 * int(x.split("_")[1]) + int(x.split("_")[-1]))
    with open(args.output_file, "w") as f:
        f.write("\n".join(val_samples))

    print(f"Saved {len(val_samples)} sample names to {args.output_file}")


if __name__ == "__main__":
    parser: ArgumentParser = _init_parser()
    args: Namespace = parser.parse_args()
    main(args)
